/*
   This client implements only PUT.

   This client takes
      hostname  port  path  file
   as command-line arguments. The
   contents of "file" are uploaded
   by the PUT method to the resource
   "path".
*/

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.*;

public class HttpClient_v5_PUT
{
   private static final int SERVER_PORT = 8080; // Should be above 1023.

   public static void main (String[] args) throws IOException
   {
      final String hostName;
      if (args.length > 0)
      {
         hostName = args[0];
      }
      else
      {
         hostName = null;
      }

      final int portNumber;
      if (args.length > 1)
      {
         portNumber = Integer.parseInt(args[1]);
      }
      else
      {
         portNumber = 0;
      }

      final String path;
      if (args.length > 2)
      {
         path = args[2];
      }
      else
      {
         path = null;
      }

      final String fileName;
      if (args.length > 3)
      {
         fileName = args[3];
      }
      else
      {
         fileName = null;
      }

      if (args.length < 3)
      {
         System.out.println("Usage: java HttpClient_v5_PUT  hostname  port  path  file");
         System.exit(-1);
      }

      // Get this client's process id number (PID). This helps
      // to identify the client in the server's transcrip.
      final ProcessHandle handle = ProcessHandle.current();
      final long pid = handle.pid();
      System.out.println("CLIENT: Process ID number (PID): " + pid );

      // Make a connection to the server.
      Socket        socket = null;
      BufferedReader    in = null;
      DataOutputStream out = null;
      try
      {
         System.out.println("CLIENT: connecting to server: " + hostName + " on port " + portNumber );
         socket = new Socket(InetAddress.getByName(hostName), portNumber);

         in = new BufferedReader(
                  new InputStreamReader(
                       socket.getInputStream()));
         out = new DataOutputStream(socket.getOutputStream());
      }
      catch (IOException e)
      {
         System.out.println("CLIENT: Cannot connect to server.");
         e.printStackTrace();
         System.exit(-1);
      }

      // Check for the existence of the file to PUT.
      final File file = new File(fileName);
      if (! file.exists())
      {
         System.out.println("CLIENT: File not found.");
      }
      if (! file.canRead())
      {
         System.out.println("CLIENT: File is not accessible.");
      }
      final int numOfBytes = (int)file.length();

      // Send the server the request headers.
             out.writeBytes("PUT " + path + " HTTP/1.1\r\n");
      System.out.println("<PUT " + path + " HTTP/1.1");
             out.writeBytes("Host: " + hostName + ":" + portNumber + "\r\n");
      System.out.println("<Host: " + hostName + ":" + portNumber);
             out.writeBytes("Connection: close\r\n");
      System.out.println("<Connection: close");
      final String contentType;
      if ( fileName.endsWith(".html") || fileName.endsWith(".htm") )
         contentType = "text/html";
      else if ( fileName.endsWith(".css") )
         contentType = "text/css";
      else if ( fileName.endsWith(".js") )
         contentType = "application/javascript";
      else if ( fileName.endsWith(".jpg") || fileName.endsWith(".jpeg") )
          contentType = "image/jpeg";
      else if ( fileName.endsWith(".gif") )
         contentType = "image/gif";
      else if ( fileName.endsWith(".png") )
          contentType = "image/png";
      else  // default content type
         contentType = "text/plain";
      out.writeBytes("Content-Type: " + contentType + "\r\n");
      System.out.println("<Content-Type: " + contentType);
      out.writeBytes("Content-Length: " + numOfBytes + "\r\n");
      System.out.println("<Content-Length: " + numOfBytes);
      // Add a blank line to denote end of request headers.
      out.writeBytes("\r\n");
      System.out.println("<");
             out.flush();

      // Send the server the request body (the file to PUT).
      try
      {
         final FileInputStream inFile = new FileInputStream(file);
         final byte[] fileInBytes = new byte[numOfBytes];
         inFile.read(fileInBytes);

         // Send the entity body.
         out.write(fileInBytes, 0, numOfBytes);
         out.flush();
         System.out.println("<=== File " + fileName + " sent to Server");
      }
      catch (FileNotFoundException e)
      {
         System.out.println("CLIENT: Unable to find file.");
         System.out.println( e );
      }
      catch (IOException e)
      {
         System.out.println("CLIENT: Unable to read file.");
         System.out.println( e );
      }


      // Receive the response from the server.
      try
      {
         String oneLine;
         // Read the response headers.
         while ( (oneLine = in.readLine()) != null )
         {
            System.out.println(">" + oneLine);
            if ( oneLine.isEmpty() )
            {
               break;
            }
         }
         // Read the entity body.
         while ( (oneLine = in.readLine()) != null )
         {
            System.out.println(oneLine);
         }
      }
      catch (IOException e)
      {
         System.out.println("CLIENT: Cannot receive response from server.");
         System.out.println( e );
      }

      // Disconnect from the server.
      try
      {
         socket.close();
         System.out.println("CLIENT: Closed socket.");
      }
      catch (IOException e)
      {
         System.out.println("CLIENT: Cannot disconnect from server.");
         System.out.println( e );
      }
   }
}
