/*
   This server takes
      port  root_directory
   as optional command-line arguments.

   This version implements persistent connections.

   There are three ways to terminate a persistent socket connection.
   1) The client socket times out when this server tries to read a request line.
   2) The client sends a null request line.
   3) The client sends a "Connection: close" request header.
*/

import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.IOException;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;
import java.net.*;
import java.util.Properties;
import java.time.format.DateTimeFormatter;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;

/**

*/
class HttpServer_v2
{
   private static final int SERVER_PORT = 8080; // Should be above 1023.
   private static final String SERVER_ROOT = "public_html";

   private final int portNumber;
   private final String serverRoot;

   /**
      This constructor intializes the server's port number
      and root folder. Then it creates the server socket
      and begins listening on it. When a client connection
      is established, it is dispatched to the handleConnection()
      method.
   */
   public HttpServer_v2(int portNumber, String serverRoot)
   {
      this.portNumber = portNumber;
      this.serverRoot = serverRoot;

      // Get this server's process id number (PID). This helps
      // to identify the server in TaskManager or TCPView.
      final ProcessHandle handle = ProcessHandle.current();
      final long pid = handle.pid();
      System.out.println("SERVER: Process ID number (PID): " + pid );

      // Get the name and IP address of the local host and
      // print them on the console for information purposes.
      try
      {
         final InetAddress address = InetAddress.getLocalHost();
         System.out.println("SERVER Hostname: " + address.getCanonicalHostName() );
         System.out.println("SERVER IP address: " +address.getHostAddress() );
         System.out.println("SERVER Using port no. " + portNumber);
      }
      catch (UnknownHostException e)
      {
         System.out.println("SERVER: Unable to determine this host's address.");
         System.out.println( e );
      }

      // Create the server's listening socket.
      ServerSocket serverSocket = null;
      try
      {
         serverSocket = new ServerSocket(portNumber);
         System.out.println("SERVER online:");
      }
      catch (IOException e)
      {
         System.out.println("SERVER: Error creating network connection.");
         e.printStackTrace(System.out);
         System.exit(-1);
      }

      int clientCounter = 0;
      while (true) // Run forever, accepting and servicing clients.
      {
         // Wait for an incoming client request.
         try
         {
            System.out.println("\nSERVER: Waiting for client " + (1+clientCounter) + " to connect.");
            final Socket clientSock = serverSocket.accept();

            ++clientCounter;
            // Log information about the connection that was just established.
            final InetAddress clientIP = clientSock.getInetAddress();
            final String clientHostAddress = clientIP.getHostAddress();
            final String clientHostName = clientIP.getCanonicalHostName();
            final int clientPortNumber = clientSock.getPort();
            System.out.println();
            System.out.println("SERVER: Client " + clientCounter + ": "
               + clientHostName + " at " + clientHostAddress + ":" + clientPortNumber);

            handleConnection(clientSock, clientCounter);
         }
         catch (SocketTimeoutException e)
         {
            System.out.println("===> Client's socket timed out!");
            e.printStackTrace();
         }
         catch (IOException e)
         {
            System.out.println("SERVER: Error communicating with client (Client no. " + clientCounter + ")");
            System.out.println( e );
         }
      }
   }


   /**
      Dispatch one client request to the appropriate handler method.
   */
   public void handleConnection(Socket clientSocket, int clientNumber)
   throws UnsupportedEncodingException, IOException
   {
      // Make the low-level input and output streams easier to use.
      BufferedReader inFromClient = new BufferedReader(
                                       new InputStreamReader(
                                          clientSocket.getInputStream()));
      PrintStream outToClient = new PrintStream(
                                   clientSocket.getOutputStream());

      boolean keepAlive = true;
      while (keepAlive)
      {
         System.out.println("===> Reading request line & request headers from Client " + clientNumber);

         //**** Read the request line.
         clientSocket.setSoTimeout(500); // timeout after 1/2 second
         String requestLine;
         try
         {
            requestLine = inFromClient.readLine();
         }
         catch (SocketTimeoutException e) // the first way to end the persistent connection
         {
            System.out.println("===> Client's socket timed out.");
            keepAlive = false;
            break;  // Finished with this client.
         }
         clientSocket.setSoTimeout(0); // remove the timer from the socket

         // Log the request line.
         System.out.println(">" + requestLine);

         if (null == requestLine) // the second way to end the persistent connection
         {
            System.out.println("===> null Request Line");
            keepAlive = false;
            break;  // Finished with this client.
         }
         requestLine = requestLine.trim();


         //**** Read the request headwrs.
         final HeaderData headerData = new HeaderData();
         headerData.connectionClose = false; // default for HTTP/1.1

         String requestHeader;
         while ( (requestHeader = inFromClient.readLine()) != null )
         {
            // Log the request header.
            System.out.println(">" + requestHeader);
            if ( requestHeader.startsWith("Connection") )
            {
               final int i = requestHeader.indexOf(':');
               final String value = requestHeader.substring(i+1).trim();
               if ( value.equals("close") )  // the third way to end the persistent connection
               {
                  headerData.connectionClose = true;
                  keepAlive = false;
               }
            }
            else if ( requestHeader.startsWith("Content-Length") )
            {
               final int i = requestHeader.indexOf(':');
               headerData.entityLength = Integer.parseInt(requestHeader.substring(i+2));
               // Log the value of Content-Length
               //System.out.println("===> Content-Length = " + headerData.entityLength);
            }
            else if ( requestHeader.isEmpty() )  // stop on a blank line
            {
               break;
            }
         }
         //**** Done reading the request headwers.


         //**** Parse and validate the request line.
         // Break the request line into a request method, url, and http version.
         final int index1 = requestLine.indexOf(" ");
         final int index2 = requestLine.lastIndexOf(" ");
         final String requestMethod = requestLine.substring(0, index1).toUpperCase();
         final String requestURL = requestLine.substring(index1, index2).trim();
         final String requestHTTP = requestLine.substring(index2).trim().toUpperCase();

         // Check for a proper version of HTTP.
         if ( ! ( requestHTTP.equals("HTTP/1.1")
               || requestHTTP.equals("HTTP/1.0")
               || requestHTTP.equals("HTTP/0.9") ) )
         {
            sendErrorResponse(outToClient,
                "505", "HTTP Version Not Supported",
                "The HTTP version in this request<br>" + requestLine +
                     "<br>is unsupported.",
                headerData,
                clientNumber);

            continue;  // Finished with this request.
         }

         // Check that the request method is supported.
         if ( ! ( requestMethod.equals("GET")
               || requestMethod.equals("HEAD")
               || requestMethod.equals("OPTIONS")
               || requestMethod.equals("POST")
               || requestMethod.equals("PUT")
               || requestMethod.equals("PATCH")
               || requestMethod.equals("DELETE")
               || requestMethod.equals("TRACE")
               || requestMethod.equals("CONNECT") ) )
         {
            sendErrorResponse(outToClient,
                "501", "Not Implemented",
                "The HTTP method in this request<br>" + requestLine +
                     "<br>is not supported.",
                headerData,
                clientNumber);

            continue;  // Finished with this request.
         }

         // Important: We need to decode the URL from its urlencoded form.
         // See
         //   http://www.w3schools.com/tags/ref_urlencode.asp
         // or
         //   http://en.wikipedia.org/wiki/URL_encoding
         final String decodedURL = URLDecoder.decode(requestURL, "UTF-8"); // throws UnsupportedEncodingException

         // Log the results of decoding the URL.
         System.out.println("===> requestURL = " + requestURL);
         System.out.println("===> decodedURL = " + decodedURL);

         // Break the decoded URL into a resource and a query.
         final String resourceName;
         final String queryString;
         if ( 0 < decodedURL.indexOf('?') )
         {
            final String[] temp = decodedURL.split("\\Q?\\E");
            resourceName  = temp[0];
            queryString   = temp[1];
         }
         else
         {
            resourceName = decodedURL;
            queryString = null;
         }

         // Important: Check that the resource is not a file
         // that "escapes" from the server's root directory.
         // This is a form of server attack that tries to
         // sneek in a URL like this.
         //    /../../../../Windows/System32/cmd.exe
         final File file1 = new File(serverRoot).getCanonicalFile();
         final File file2 = new File(serverRoot + resourceName).getCanonicalFile();
         if (! file2.getAbsolutePath().startsWith(file1.getAbsolutePath()))
         {
            sendErrorResponse(outToClient,
                "403", "Forbidden",
                "You can't pull this stunt<br>"+requestLine+"<br>on us.",
                headerData,
                clientNumber);

            keepAlive = false;
            break;  // Finished with this client.
         }

         // Log the results of parsing the request line.
         System.out.println("===> Request method = " + requestMethod);
         System.out.println("===> Resource name = " + resourceName);
         System.out.println("===> Query string = " + queryString);
         System.out.println("===> HTTP version = " + requestHTTP);
         //**** Done parsing the request line.


         //**** Dispatch the request method.
         if ( requestMethod.equals("GET")
           || requestMethod.equals("HEAD") )
         {
            doGet(requestMethod,
                  inFromClient,
                  outToClient,
                  resourceName,
                  queryString,
                  headerData,
                  clientNumber);
         }
         else if ( requestMethod.equals("OPTIONS") )
         {
            doOptions(inFromClient,
                      outToClient,
                      resourceName,
                      queryString,
                      headerData,
                      clientNumber);
         }
         else if ( requestMethod.equals("POST") )
         {
            doPost(inFromClient,
                   outToClient,
                   resourceName,
                   queryString,
                   headerData,
                   clientNumber);
         }
         else if ( requestMethod.equals("PUT") )
         {
            doPut(inFromClient,
                  outToClient,
                  resourceName,
                  queryString,
                  headerData,
                  clientNumber);
         }
         else if ( requestMethod.equals("PATCH") )
         {
            doPut(inFromClient,
                  outToClient,
                  resourceName,
                  queryString,
                  headerData,
                  clientNumber);
         }
         else if ( requestMethod.equals("DELETE") )
         {
            doDelete(inFromClient,
                     outToClient,
                     resourceName,
                     queryString,
                     headerData,
                     clientNumber);
         }
         else if ( requestMethod.equals("TRACE") )
         {
            doTrace(inFromClient,
                    outToClient,
                    resourceName,
                    queryString,
                    headerData,
                    clientNumber);
         }
         else if ( requestMethod.equals("CONNECT") )
         {
            doConnect(inFromClient,
                      outToClient,
                      resourceName,
                      queryString,
                      headerData,
                      clientNumber);
         }
         else  // unsupported request method
         {
            sendErrorResponse(outToClient,
                "501", "Not Implemented",
                "This server does not implement this method: " + requestMethod,
                headerData,
                clientNumber);
         }
      }// Continue to the next request on this socket.

      clientSocket.close();
      System.out.println("SERVER: Client " + clientNumber + ": Closed socket.");
      return;  // Finished with this client.
   }


   /**
      Dispatch static and dynamic GET requests.
   */
   public void doGet(String requestMethod,
                     BufferedReader inFromClient,
                     PrintStream outToClient,
                     String resourceName,
                     String queryString,
                     HeaderData headerData,
                     int clientNumber)
   {
      // Make sure the resource exits and is readable.
      final File file = new File(serverRoot + resourceName);
      if ( ! file.exists() || ! file.canRead() )
      {
         if (! file.exists())
         {
            sendErrorResponse(outToClient,
                "404", "Not Found",
                "The requested URL "+resourceName+" was not found on this server.",
                headerData,
                clientNumber);
         }
         else
         {
            sendErrorResponse(outToClient,
                "403", "Forbidden",
                "The requested URL "+resourceName+" is not readable.",
                headerData,
                clientNumber);
         }

         return;  // Finished with this request.
      }

      if ( file.isDirectory() )
      {
         // Check for existence of an index.html file.
         final File file2 = new File(serverRoot + resourceName + "index.html");
         if ( file2.exists() && file2.canRead() )
         {
            doStaticGet(requestMethod,
                        inFromClient,
                        outToClient,
                        resourceName + "index.html",
                        queryString,
                        headerData,
                        clientNumber,
                        file2);
         }
         else
         {
            doDynDirListing(requestMethod,
                            inFromClient,
                            outToClient,
                            resourceName,
                            queryString,
                            headerData,
                            clientNumber,
                            file);
         }
      }
      else if ( resourceName.endsWith(".class")
             || resourceName.endsWith(".exe") )
      {
         doCGIGet(requestMethod,
                  inFromClient,
                  outToClient,
                  resourceName,
                  queryString,
                  headerData,
                  clientNumber,
                  file);
      }
      else // the file exists, so send it to the client
      {
         doStaticGet(requestMethod,
                     inFromClient,
                     outToClient,
                     resourceName,
                     queryString,
                     headerData,
                     clientNumber,
                     file);
      }
   }



   /**
      Send a static resource to the client.
   */
   public void doStaticGet(String requestMethod,
                           BufferedReader inFromClient,
                           PrintStream outToClient,
                           String resourceName,
                           String queryString,
                           HeaderData headerData,
                           int clientNumber,
                           File file)
   {
      final int numOfBytes = (int)file.length();
      try
      {
         final FileInputStream inFile = new FileInputStream(file);
         final byte[] fileInBytes = new byte[numOfBytes];
         inFile.read(fileInBytes);

         // Send the HTTP response line and headers.
         // The response line.
         outToClient.print("HTTP/1.1 200 OK\r\n");
         System.out.println("<HTTP/1.1 200 OK");
         // The response headers.
         final String date = OffsetDateTime.now(ZoneOffset.UTC)
                               .format(DateTimeFormatter.RFC_1123_DATE_TIME);
         outToClient.print("Date: " + date + "\r\n");
         System.out.println("<Date: " + date);
         outToClient.print("Server: HttpServer_v2.java\r\n");
         System.out.println("<Server: HttpServer_v2.java");
         if (headerData.connectionClose)
         {
            outToClient.print("Connection: close\r\n");
            System.out.println("<Connection: close");
         }
         final String contentType;
         if ( resourceName.endsWith(".html") || resourceName.endsWith(".htm") )
            contentType = "text/html";
         else if ( resourceName.endsWith(".css") )
            contentType = "text/css";
         else if ( resourceName.endsWith(".js") )
            contentType = "application/javascript";
         else if ( resourceName.endsWith(".jpg") || resourceName.endsWith(".jpeg") )
             contentType = "image/jpeg";
         else if ( resourceName.endsWith(".gif") )
            contentType = "image/gif";
         else if ( resourceName.endsWith(".png") )
             contentType = "image/png";
         else  // default content type
            contentType = "text/plain";
         outToClient.print("Content-Type: " + contentType + "\r\n");
         System.out.println("<Content-Type: " + contentType);
         outToClient.print("Content-Length: " + numOfBytes + "\r\n");
         System.out.println("<Content-Length: " + numOfBytes);
         // Add a blank line to denote end of response headers.
         outToClient.print("\r\n");
         System.out.println("<");

         // Send the entity body.
         if ( requestMethod.equals("GET") ) // not HEAD
         {
            outToClient.write(fileInBytes, 0, numOfBytes);
            System.out.println("<=== Resource " + resourceName + " sent to Client " + clientNumber);
         }
         outToClient.flush();
      }
      catch (FileNotFoundException e)
      {
         System.out.println("SERVER: doStaticGet: Unable to find file.");
         System.out.println( e );
      }
      catch (IOException e)
      {
         System.out.println("SERVER: doStaticGet: Unable to read file.");
         System.out.println( e );
      }
   }


   /**
      This method should implement the CGI GET protocol for
      inter-process communication between the server process,
      its child process, and the client process.
   */
   public void doCGIGet(String requestMethod,
                        BufferedReader inFromClient,
                        PrintStream outToClient,
                        String resourceName,
                        String queryString,
                        HeaderData headerData,
                        int clientNumber,
                        File file)
   {
      // This is just a sketch of an implementation of the cgi GET method.
      //
      // This method should run the resource program named resourceName
      // and pass it the querryString data. The resource program would
      // then complete the transaction with the client.
      //
      // Instead, this sketch just logs the querryString to the server's
      // stdout and then passes the rest of the request to doStaticGet.
      // From the client's perspective, a cgi GET is the same as a
      // static GET.

      // Log the query string to stdout.
      System.out.println( "===> Received a CGI GET request with the following query string.\n" );
      System.out.println( "===> " + queryString + "\n" );
      System.out.println( "===> Processing as a normal GET request.\n" );

      // Send the client the static resourse.
      doStaticGet(requestMethod,
                  inFromClient,
                  outToClient,
                  resourceName,
                  queryString,
                  headerData,
                  clientNumber,
                  file);
   }


   /**
      This method should be implemented to send the client a
      html page that is a listing of the given directory.
   */
   public void doDynDirListing(String requestMethod,
                               BufferedReader inFromClient,
                               PrintStream outToClient,
                               String resourceName,
                               String queryString,
                               HeaderData headerData,
                               int clientNumber,
                               File file)
   {
      sendErrorResponse(outToClient,
          "400", "Bad Request",
          "The requested URL, "+resourceName+", does not contain a file name.",
          headerData,
          clientNumber);
   }


   /**
      Dispatch the different kinds of POST requests.
   */
   public void doPost(BufferedReader inFromClient,
                      PrintStream outToClient,
                      String resourceName,
                      String queryString,
                      HeaderData headerData,
                      int clientNumber)
   {
      // Make sure the resource exits and is readable.
      final File file = new File(serverRoot + resourceName);
      if ( ! file.exists() || ! file.canRead() )
      {
         if (! file.exists())
         {
            sendErrorResponse(outToClient,
                "404", "Not Found",
                "The requested URL "+resourceName+" was not found on this server.",
                headerData,
                clientNumber);
         }
         else
         {
            sendErrorResponse(outToClient,
                "403", "Forbidden",
                "The requested URL "+resourceName+" is not readable.",
                headerData,
                clientNumber);
         }

         return;  // Finished with this request.
      }

      if ( file.isDirectory() )
      {
         sendErrorResponse(outToClient,
             "405", "Method Not Allowed",
             "The POST method applied to a directory<br>" + resourceName +
                   "<br>is not allowed.",
             headerData,
             clientNumber);

         return;  // Finished with this request.
      }

      // Read the entity body.
      final int entityLength = headerData.entityLength;
      final char[] entityBodyChars = new char[entityLength];
      int totalCharsRcvd = 0;  // Total chars received so far.
      int charsRcvd;           // chars received in last read.
      while ( totalCharsRcvd < entityLength )
      {
         System.out.println("===> Reading entity body.");
         try
         {
            charsRcvd = inFromClient.read(entityBodyChars,
                                          totalCharsRcvd,
                                          entityLength - totalCharsRcvd);
            if (-1 == charsRcvd)
            {
               break; // reached end-of-file
            }
            totalCharsRcvd += charsRcvd;
         }
         catch (IOException e)
         {
            System.out.println("SERVER: doPost: Unable to read entity body.");
            System.out.println( e );
         }
         System.out.println("===> totalCharsRcvd = " + totalCharsRcvd);
      }

      if ( resourceName.endsWith(".class")
        || resourceName.endsWith(".exe") )
      {
         doCGIPost(inFromClient,
                   outToClient,
                   resourceName,
                   queryString,
                   headerData,
                   clientNumber,
                   file,
                   entityBodyChars);
      }
      else if ( resourceName.equals("/post_data") )
      {
         doBuiltInPost(inFromClient,
                       outToClient,
                       resourceName,
                       queryString,
                       headerData,
                       clientNumber,
                       file,
                       entityBodyChars);
      }
      else
      {
         doStaticPost(inFromClient,
                      outToClient,
                      resourceName,
                      queryString,
                      headerData,
                      clientNumber,
                      file,
                      entityBodyChars);
      }
   }


   /**
      This method should implement the CGI POST protocol for
      inter-process communication between the server process,
      its child process, and the client.
   */
   public void doCGIPost(BufferedReader inFromClient,
                         PrintStream outToClient,
                         String resourceName,
                         String queryString,
                         HeaderData headerData,
                         int clientNumber,
                         File file,
                         char[] entityBodyChars)
   {
      // Log the entity body to stdout.
      final String entityBody = new String(entityBodyChars);
      System.out.println( "===> " + entityBody + "\n" );

      // We need to launch the Java program called resourceName.
      // Get the path of the java.exe program that started this server's JVM.
      String javaPath = ProcessHandle.current()
                        .info()
                        .command()
                        .orElse(null);
      if(javaPath == null)
      {
         System.out.println("SERVER: Could not get the path to java.exe.");
         return;
      }
      System.out.println("SERVER: Using this JVM for POST: " + javaPath);

      // Configure the ProcessBuilder.
      try
      {
         final String resourceClass = resourceName
                                     .substring(0, resourceName.lastIndexOf("."));
         final String classFile = new File(serverRoot + resourceClass)
                                      .getCanonicalFile()
                                      .getAbsolutePath();
         System.out.println("SERVER: Running this class file: " + classFile);

         final ProcessBuilder pb = new ProcessBuilder(
                                         javaPath,
                                         "-cp",
                                         new File(serverRoot).getCanonicalFile().getAbsolutePath(),
                                         resourceClass.substring(1));
         System.out.println( pb.command() );

         // Start a new java process.
         final Process p = pb.start();
         System.out.println("SERVER: Launced this child process: " + p);
         // Write POST data to the input stream of the process.
         try (PrintStream out = new PrintStream(
                                   p.getOutputStream()))
         {
            out.print(entityBodyChars);
            out.flush();
         }

         // Send the HTTP response line and some headers.
         // The response line.
         outToClient.print("HTTP/1.1 200 OK\r\n");
         System.out.println("<HTTP/1.1 200 OK");
         // The first response headers.
         final String date = OffsetDateTime.now(ZoneOffset.UTC)
                               .format(DateTimeFormatter.RFC_1123_DATE_TIME);
         outToClient.print("Date: " + date + "\r\n");
         System.out.println("<Date: " + date);
         outToClient.print("Server: HttpServer_v2.java\r\n");
         System.out.println("<Server: HttpServer_v2.java");
         if (headerData.connectionClose)
         {
            outToClient.print("Connection: close\r\n");
            System.out.println("<Connection: close");
         }

         // Read the standard output stream of the process
         // and forward the results to the client.
         System.out.println("====> Server reading response from child process.");
         try (BufferedReader in = new BufferedReader(
                                    new InputStreamReader(
                                      p.getInputStream())))
         {
            String oneLine;
            while ((oneLine = in.readLine()) != null)
            {
               outToClient.println(oneLine);
               System.out.println("<" + oneLine);
            }
         }
         outToClient.flush();
         System.out.println("====> Server done reading response from child process.");

         // Read the standard error stream of the process.
         System.out.println("====> Server reading error stream from child process.");
         try (BufferedReader err = new BufferedReader(
                                     new InputStreamReader(
                                       p.getErrorStream())))
         {
            String oneLine;
            while ((oneLine = err.readLine()) != null)
            {
               System.out.println("err: " + oneLine);
            }
         }
         System.out.println("====> Server done reading errors from child process.");
      }
      catch (IOException e)
      {
         System.out.println("SERVER: Error while runing " + resourceName);
         System.out.println( e );
      }
   }


   /**
      This method implements a simple built-in POST protocol.
      This method prints the entity body to the server's stdout
      and then sends back to the client a dynamically generated
      "thank you" web page which includes a copy of the client's
      POST data.

      In general, a method like this can do any amount of
      computation using the client's POST data as the input
      to the calculations. And the returned web page can
      then depend on those calculations.
   */
   public void doBuiltInPost(BufferedReader inFromClient,
                             PrintStream outToClient,
                             String resourceName,
                             String queryString,
                             HeaderData headerData,
                             int clientNumber,
                             File file,
                             char[] entityBodyChars)
   {
      // Log the entity body to stdout.
      final String entityBody = new String(entityBodyChars);
      System.out.println( "===> " + entityBody + "\n" );

      // Build the HTTP response body.
      final String body = "<!doctype html>\r\n" +
                          "<html>\r\n" +
                          "<head>\r\n" +
                          "<title>Thank you!</title>\r\n" +
                          "</head>\r\n" +
                          "<body>\r\n" +
                          "<h1>Thank you!</h1>\r\n" +
                          "<p>Thank you for your POST data!</p>\r\n" +
                          "<p></p>\r\n" +
                          "<p>" + entityBody + "</p>\r\n" +
                          "<p></p>\r\n" +
                          "<p>We will protect it forever.</p>\r\n" +
                          "</body>\r\n" +
                          "</html>\r\n";

      // Send the HTTP response line and headers.
      // The response line.
      outToClient.print("HTTP/1.1 200 OK\r\n");
      System.out.println("<HTTP/1.1 200 OK");
      // The response headers.
      final String date = OffsetDateTime.now(ZoneOffset.UTC)
                            .format(DateTimeFormatter.RFC_1123_DATE_TIME);
      outToClient.print("Date: " + date + "\r\n");
      System.out.println("<Date: " + date);
      outToClient.print("Server: HttpServer_v2.java\r\n");
      System.out.println("<Server: HttpServer_v2.java");
      if (headerData.connectionClose)
      {
         outToClient.print("Connection: close\r\n");
         System.out.println("<Connection: close");
      }
      outToClient.print("Content-Type: text/html\r\n");
      System.out.println("<Content-Type: text/html");
      outToClient.print("Content-Length: " + body.length() + "\r\n");
      System.out.println("<Content-Length: " + body.length());
      // Add a blank line to denote end of response headers.
      outToClient.print("\r\n");
      System.out.println("<");

      // Send the entity body.
      outToClient.print(body);
      outToClient.flush();
      System.out.println("<=== Entity body sent to Client " + clientNumber);
   }


   /**
      This method should implement a static POST protocol.
   */
   public void doStaticPost(BufferedReader inFromClient,
                            PrintStream outToClient,
                            String resourceName,
                            String queryString,
                            HeaderData headerData,
                            int clientNumber,
                            File file,
                            char[] entityBodyChars)
   {
      // This is just a sketch of an implementation of the POST method.
      // This sketch prints the entity body to the server's stdout and
      // then sends back to the client the static resource resourceName.

      // Log the entity body to stdout.
      final String data = new String(entityBodyChars);
      System.out.println( "===> " + data + "\n" );

      // Note: We could substitute the entity data string
      //       for the queryString and then call doCGIGet().

      // Send the client the static resourse.
      doStaticGet("GET",
                  inFromClient,
                  outToClient,
                  resourceName,
                  queryString,
                  headerData,
                  clientNumber,
                  file);
   }


   /**
      This method should implement the HTTP OPTIONS command.
   */
   public void doOptions(BufferedReader inFromClient,
                         PrintStream outToClient,
                         String resourceName,
                         String queryString,
                         HeaderData headerData,
                         int clientNumber)
   {
      // Send the HTTP response line and headers.
      // The response line.
      outToClient.print("HTTP/1.1 200 OK\r\n");
      System.out.println("<HTTP/1.1 200 OK");
      // The response headers.
      final String date = OffsetDateTime.now(ZoneOffset.UTC)
                            .format(DateTimeFormatter.RFC_1123_DATE_TIME);
      outToClient.print("Date: " + date + "\r\n");
      System.out.println("<Date: " + date);
      outToClient.print("Server: HttpServer_v2.java\r\n");
      System.out.println("<Server: HttpServer_v2.java");
      outToClient.print("Allow: GET,HEAD,POST,OPTIONS\r\n");
      System.out.println("<Allow: GET,HEAD,POST,OPTIONS");
      if (headerData.connectionClose)
      {
         outToClient.print("Connection: close\r\n");
         System.out.println("<Connection: close");
      }
      outToClient.print("Content-type: text/html\r\n");
      System.out.println("<Content-type: text/html");
      outToClient.print("Content-Length: 0 \r\n");
      System.out.println("<Content-Length: 0");
      // Add a blank line to denote end of response headers.
      outToClient.print("\r\n");
      System.out.println("<");
      outToClient.flush();
   }


   /**
      This method should implement the HTTP PUT command.
   */
   public void doPut(BufferedReader inFromClient,
                     PrintStream outToClient,
                     String resourceName,
                     String queryString,
                     HeaderData headerData,
                     int clientNumber)
   {
      // Read the entity body.
      final int entityLength = headerData.entityLength;
      final char[] entityBodyChars = new char[entityLength];
      int totalCharsRcvd = 0;  // Total chars received so far.
      int charsRcvd;           // chars received in last read.
      while ( totalCharsRcvd < entityLength )
      {
         System.out.println("===> Reading entity body.");
         try
         {
            charsRcvd = inFromClient.read(entityBodyChars,
                                          totalCharsRcvd,
                                          entityLength - totalCharsRcvd);
            if (-1 == charsRcvd)
            {
               break; // reached end-of-file
            }
            totalCharsRcvd += charsRcvd;
         }
         catch (IOException e)
         {
            System.out.println("SERVER: doPut: Unable to read entity body.");
            System.out.println( e );
         }
         System.out.println("===> totalCharsRcvd = " + totalCharsRcvd);
      }

      // Store the entity body in the server's file system.
      final File file = new File(serverRoot + "/files/" + resourceName);
      PrintWriter fileOut = null;
      try
      {
         fileOut = new PrintWriter(file);
      }
      catch (FileNotFoundException e)
      {
            System.out.println("SERVER: doPut: Unable to write destination file.");
            System.out.println( e );
      }
      fileOut.print(entityBodyChars);
      fileOut.flush();
      fileOut.close();
      //System.out.print(entityBodyChars);

      // Send the HTTP response line and headers.
      // The response line.
      outToClient.print("HTTP/1.1 201 Created\r\n");
      System.out.println("<HTTP/1.1 201 Created");
      // The response headers.
      final String date = OffsetDateTime.now(ZoneOffset.UTC)
                            .format(DateTimeFormatter.RFC_1123_DATE_TIME);
      outToClient.print("Date: " + date + "\r\n");
      System.out.println("<Date: " + date);
      outToClient.print("Server: HttpServer_v2.java\r\n");
      System.out.println("<Server: HttpServer_v2.java");
      outToClient.print("Connection: close\r\n");
      System.out.println("<Connection: close");
      outToClient.print("Content-type: text/html\r\n");
      System.out.println("<Content-type: text/html");
      outToClient.print("Content-Length: 0 \r\n");
      System.out.println("<Content-Length: 0");
      // Add a blank line to denote end of response headers.
      outToClient.print("\r\n");
      System.out.println("<");
      outToClient.flush();
   }


   /**
      This method should implement the HTTP PATCH command.
   */
   public void doPatch(BufferedReader inFromClient,
                       PrintStream outToClient,
                       String resourceName,
                       String queryString,
                       HeaderData headerData,
                       int clientNumber)
   {
      sendErrorResponse(outToClient,
          "400", "Bad Request",
          "This server cannot yet handle PATCH requests.",
          headerData,
          clientNumber);
   }


   /**
      This method should implement the HTTP DELETE command.
   */
   public void doDelete(BufferedReader inFromClient,
                        PrintStream outToClient,
                        String resourceName,
                        String queryString,
                        HeaderData headerData,
                        int clientNumber)
   {
      sendErrorResponse(outToClient,
          "400", "Bad Request",
          "This server cannot yet handle DELETE requests.",
          headerData,
          clientNumber);
   }


   /**
      This method should implement the HTTP TRACE command.
   */
   public void doTrace(BufferedReader inFromClient,
                       PrintStream outToClient,
                       String resourceName,
                       String queryString,
                       HeaderData headerData,
                       int clientNumber)
   {
      sendErrorResponse(outToClient,
          "400", "Bad Request",
          "This server cannot yet handle TRACE requests.",
          headerData,
          clientNumber);
   }


   /**
      This method should implement the HTTP CONNECT command.
   */
   public void doConnect(BufferedReader inFromClient,
                         PrintStream outToClient,
                         String resourceName,
                         String queryString,
                         HeaderData headerData,
                         int clientNumber)
   {
      sendErrorResponse(outToClient,
          "400", "Bad Request",
          "This server cannot yet handle CONNECT requests.",
          headerData,
          clientNumber);
   }


   /**

   */
   public void sendErrorResponse(PrintStream outToClient,
                                 String code,
                                 String shortMsg,
                                 String longMsg,
                                 HeaderData headerData,
                                 int clientNumber)
   {
      String localHostName = null;
      try
      {
         localHostName = InetAddress.getLocalHost().getCanonicalHostName();
      }
      catch (UnknownHostException e)
      {
         System.out.println("SERVER: Unable to determine this host's name.");
         System.out.println( e );
      }

      // Build the response body.
      String body = "<!doctype html>\r\n" +
                    "<html>\r\n" +
                    "<head>\r\n" +
                    "<title>" + code + " " + shortMsg + "</title>\r\n" +
                    "</head>\r\n" +
                    "<body>\r\n" +
                    "<h1>" + shortMsg + "</h1>\r\n" +
                    "<p>" + code + ": " + shortMsg + "</p>\r\n" +
                    "<p>" + longMsg + "</p>\r\n" +
                    "<hr>\r\n" +
                    "<address>HttpServer_v2.java at " +
                        localHostName + " Port " + portNumber +
                    "</address>\r\n" +
                    "</body>\r\n" +
                    "</html>\r\n";

      // Send the HTTP response line and headers.
      // The response line.
      outToClient.print("HTTP/1.1 " + code + " " + shortMsg + "\r\n");
      System.out.println("<HTTP/1.1 " + code + " " + shortMsg);
      // The response headers.
      final String date = OffsetDateTime.now(ZoneOffset.UTC)
                            .format(DateTimeFormatter.RFC_1123_DATE_TIME);
      outToClient.print("Date: " + date + "\r\n");
      System.out.println("<Date: " + date);
      outToClient.print("Server: HttpServer_v2.java\r\n");
      System.out.println("<Server: HttpServer_v2.java");
      if (headerData.connectionClose)
      {
         outToClient.print("Connection: close\r\n");
         System.out.println("<Connection: close");
      }
      outToClient.print("Content-Type: text/html\r\n");
      System.out.println("<Content-Type: text/html");
      outToClient.print("Content-Length: " + body.length() + "\r\n");
      System.out.println("<Content-Length: " + body.length());
      // Add a blank line to denote end of response headers.
      outToClient.print("\r\n");
      System.out.println("<");

      // Send the entity body.
      outToClient.print(body);
      outToClient.flush();
      System.out.println("<=== Entity body sent to Client " + clientNumber);
   }


   private static class HeaderData
   {
      public boolean connectionClose;
      public boolean connectionKeepAlive;
      public int keepAliveTimeout;
      public int keepAliveMax;
      public int entityLength;
   }


   /**
      The main() method configures the server using
      a httpserver.properties file and command-line
      arguments.
   */
   public static void main(String[] args)
   {
      String serverRoot = SERVER_ROOT;   // default value
      int portNumber = SERVER_PORT;      // default value

      final Properties properties = new Properties();
      try
      {
         properties.load(
            new FileInputStream(
               new File("httpserver.properties")));

         serverRoot = properties.getProperty("root");
         final String port = properties.getProperty("port");
         if (port != null) portNumber = Integer.parseInt(port);
      }
      catch (FileNotFoundException e)
      {
         /* ignore FileNotFoundException */
      }
      catch (IOException e)
      {
         e.printStackTrace(System.out);
         System.exit(-1);
      }

      if (0 < args.length)
      {
         portNumber = Integer.parseInt(args[0]);
      }
      if (1 < args.length)
      {
         serverRoot = args[1];
      }

      // Instantiate and start up an instance of the server.
      new HttpServer_v2(portNumber, serverRoot);
   }
}
